#!/usr/bin/env python
# License: GPLv3 Copyright: 2026, calibre contributors

from calibre.srv.routes import endpoint
from calibre.srv.tests.base import BaseTest


class TestContentServerPlugin(BaseTest):
    def test_basic_endpoint_registration(self):
        from calibre.customize import ContentServerPlugin

        class TestPlugin(ContentServerPlugin):
            name = 'Test Content Plugin'

            def content_server_endpoints(self):
                @endpoint('/test-plugin/hello', auth_required=False)
                def hello(ctx, rd):
                    return 'hello'

                return [hello]

        plugin = TestPlugin(None)
        endpoints = list(plugin.content_server_endpoints())
        self.assertEqual(len(endpoints), 1)
        self.assertTrue(getattr(endpoints[0], 'is_endpoint', False))
        self.assertEqual(endpoints[0].route, '/test-plugin/hello')
        self.assertFalse(endpoints[0].auth_required)

    def test_default_empty_endpoints(self):
        from calibre.customize import ContentServerPlugin

        plugin = ContentServerPlugin(None)
        endpoints = plugin.content_server_endpoints()
        self.assertEqual(len(list(endpoints)), 0)

    def test_multiple_endpoints(self):
        from calibre.customize import ContentServerPlugin

        class MultiPlugin(ContentServerPlugin):
            name = 'Multi Route Plugin'

            def content_server_endpoints(self):
                @endpoint('/multi/a')
                def a(ctx, rd):
                    pass

                @endpoint('/multi/b', auth_required=False)
                def b(ctx, rd):
                    pass

                @endpoint('/multi/{x}', types={'x': int})
                def c(ctx, rd, x):
                    pass

                return [a, b, c]

        plugin = MultiPlugin(None)
        eps = list(plugin.content_server_endpoints())
        self.assertEqual(len(eps), 3)
        self.assertTrue(eps[0].auth_required)
        self.assertFalse(eps[1].auth_required)
        self.assertEqual(eps[2].types, {'x': int})


class TestPluginDiscovery(BaseTest):
    """
    Test that ContentServerPlugin endpoints are discovered and registered.
    Uses Router directly rather than Handler to avoid PyQt6 dependency
    in the test environment (calibre.srv.ajax -> calibre.ebooks.covers -> QtGui).
    """

    def setUp(self):
        from calibre.customize.ui import _initialized_plugins, config

        self._orig_plugins = list(_initialized_plugins)
        self._orig_disabled = set(config['disabled_plugins'])

    def tearDown(self):
        from calibre.customize.ui import _initialized_plugins, config

        _initialized_plugins[:] = self._orig_plugins
        config['disabled_plugins'] = self._orig_disabled

    def _register_all_plugin_routes(self, router):
        from calibre.customize.ui import content_server_plugins

        for plugin in content_server_plugins():
            for ep in plugin.content_server_endpoints():
                try:
                    router.add(ep)
                except ValueError:
                    pass

    def test_plugin_routes_loaded(self):
        from calibre.customize import ContentServerPlugin
        from calibre.customize.ui import _initialized_plugins
        from calibre.srv.routes import Router

        class RoutePlugin(ContentServerPlugin):
            name = 'Test Route Plugin'

            def content_server_endpoints(self):
                @endpoint('/test-route/ping', auth_required=False)
                def ping(ctx, rd):
                    return 'pong'

                return [ping]

        plugin = RoutePlugin(None)
        _initialized_plugins.append(plugin)
        router = Router()
        self._register_all_plugin_routes(router)
        router.finalize()
        path = tuple(filter(None, '/test-route/ping'.split('/')))
        ep, args = router.find_route(path)
        self.assertEqual(ep.route, '/test-route/ping')
        self.assertFalse(ep.auth_required)
        _initialized_plugins.remove(plugin)

    def test_disabled_plugin_skipped(self):
        from calibre.customize import ContentServerPlugin
        from calibre.customize.ui import _initialized_plugins, config
        from calibre.srv.routes import HTTPNotFound, Router

        class RoutePlugin(ContentServerPlugin):
            name = 'Disabled Route Plugin'

            def content_server_endpoints(self):
                @endpoint('/disabled-route/test', auth_required=False)
                def test(ctx, rd):
                    return 'test'

                return [test]

        plugin = RoutePlugin(None)
        _initialized_plugins.append(plugin)
        config['disabled_plugins'].add(plugin.name)
        router = Router()
        self._register_all_plugin_routes(router)
        router.finalize()
        path = tuple(filter(None, '/disabled-route/test'.split('/')))
        with self.assertRaises(HTTPNotFound):
            router.find_route(path)
        _initialized_plugins.remove(plugin)

    def test_plugin_conflict_does_not_override(self):
        from calibre.customize import ContentServerPlugin
        from calibre.customize.ui import _initialized_plugins
        from calibre.srv.routes import Router

        @endpoint('/test-route/ping', auth_required=False)
        def original(ctx, rd):
            return 'original'

        class OverridePlugin(ContentServerPlugin):
            name = 'Override Plugin'

            def content_server_endpoints(self):
                @endpoint('/test-route/ping', auth_required=False)
                def duplicate(ctx, rd):
                    return 'duplicate'

                return [duplicate]

        plugin = OverridePlugin(None)
        _initialized_plugins.append(plugin)
        router = Router()
        router.add(original)
        self._register_all_plugin_routes(router)
        router.finalize()
        path = tuple(filter(None, '/test-route/ping'.split('/')))
        ep, args = router.find_route(path)
        self.assertEqual(ep.__name__, 'original')
        _initialized_plugins.remove(plugin)

    def test_broken_plugin_does_not_block_others(self):
        from calibre.customize import ContentServerPlugin
        from calibre.customize.ui import _initialized_plugins
        from calibre.srv.routes import Router

        class GoodPlugin(ContentServerPlugin):
            name = 'Good Plugin'

            def content_server_endpoints(self):
                @endpoint('/good/hello', auth_required=False)
                def hello(ctx, rd):
                    return 'hello'

                return [hello]

        class BrokenPlugin(ContentServerPlugin):
            name = 'Broken Plugin'

            def content_server_endpoints(self):
                raise RuntimeError('oops')

        good = GoodPlugin(None)
        broken = BrokenPlugin(None)
        _initialized_plugins.append(good)
        _initialized_plugins.append(broken)
        router = Router()
        from calibre.customize.ui import content_server_plugins

        for plugin in content_server_plugins():
            try:
                endpoints = plugin.content_server_endpoints()
            except Exception:
                continue
            for ep in endpoints:
                try:
                    router.add(ep)
                except ValueError:
                    pass
        router.finalize()
        path = tuple(filter(None, '/good/hello'.split('/')))
        ep, args = router.find_route(path)
        self.assertEqual(ep.route, '/good/hello')
        _initialized_plugins.remove(good)
        _initialized_plugins.remove(broken)



#!/usr/bin/env python
# License: GPLv3 Copyright: 2013, Kovid Goyal <kovid at kovidgoyal.net>

import copy
from collections import OrderedDict
from functools import partial

from calibre.ebooks.metadata import author_to_author_sort
from calibre.utils.config_base import prefs, tweaks
from calibre.utils.icu import collation_order, sort_key
from calibre.utils.icu import lower as icu_lower
from calibre.utils.icu import upper as icu_upper

CATEGORY_SORTS = ('name', 'popularity', 'rating')  # This has to be a tuple not a set


class Tag:
    __slots__ = (
        'avg_rating',
        'category',
        'count',
        'id',
        'id_set',
        'is_editable',
        'is_hierarchical',
        'is_searchable',
        'name',
        'original_categories',
        'original_name',
        'search_expression',
        'sort',
        'state',
        'use_sort_as_name',
    )

    def __init__(
        self,
        name,
        id=None,
        count=0,
        state=0,
        avg=0,
        sort=None,
        category=None,
        id_set=None,
        search_expression=None,
        is_editable=True,
        is_searchable=True,
        use_sort_as_name=False,
        original_categories=None,
    ):
        self.name = self.original_name = name
        self.id = id
        self.count = count
        self.state = state
        self.is_hierarchical = ''
        self.is_editable = is_editable
        self.is_searchable = is_searchable
        self.id_set = id_set if id_set is not None else set()
        self.avg_rating = avg / 2.0 if avg is not None else 0
        self.sort = sort
        self.use_sort_as_name = use_sort_as_name
        self.category = category
        self.search_expression = search_expression
        self.original_categories = None

    @property
    def string_representation(self):
        return f'{self.name}:{self.count}:{self.id}:{self.state}:{self.category}:{self.original_categories}'

    def __str__(self):
        return self.string_representation

    def __repr__(self):
        return str(self)

    __calibre_serializable__ = True

    def as_dict(self):
        return {k: getattr(self, k) for k in self.__slots__}

    @classmethod
    def from_dict(cls, d):
        ans = cls('')
        for k in cls.__slots__:
            setattr(ans, k, d[k])
        return ans


def find_categories(field_metadata):
    for category, cat in field_metadata.iter_items():
        if cat['is_category'] and cat['kind'] not in {'user', 'search'}:
            yield (category, cat['is_multiple'].get('cache_to_list', None), False)
        elif cat['datatype'] == 'composite' and cat['display'].get('make_category', False):
            yield (category, cat['is_multiple'].get('cache_to_list', None), True)


def create_tag_class(category, fm):
    cat = fm[category]
    dt = cat['datatype']
    is_editable = category not in {'news', 'rating', 'languages', 'formats', 'identifiers'} and dt != 'composite'

    if (
        (category == 'authors' or (cat['display'].get('is_names', False) and cat['is_custom'] and cat['is_multiple'] and dt == 'text'))
        and tweaks['categories_use_field_for_author_name'] == 'author_sort'
    ) or (dt == 'series' and tweaks['categories_use_field_for_series_name'] == 'series_sort'):
        use_sort_as_name = True
    else:
        use_sort_as_name = False

    return partial(Tag, use_sort_as_name=use_sort_as_name, is_editable=is_editable, category=category)


def clean_user_categories(dbcache):
    user_cats = dbcache.pref('user_categories', {})
    new_cats = {}
    for k in user_cats:
        comps = [c.strip() for c in k.split('.') if c.strip()]
        if len(comps) == 0:
            i = 1
            while True:
                if str(i) not in user_cats:
                    new_cats[str(i)] = user_cats[k]
                    break
                i += 1
        else:
            new_cats['.'.join(comps)] = user_cats[k]
    try:
        if new_cats != user_cats:
            dbcache.set_pref('user_categories', new_cats)
    except Exception:
        pass
    return new_cats


def is_standard_category(key):
    return not (key.startswith('@') or key == 'search')


def category_display_order(ordered_cats, all_cats):
    # ordered_cats is the desired order. all_cats is the list of keys returned
    # by get_categories, which is in the default order
    cat_ord = []
    all_cat_set = frozenset(all_cats)
    # Do the standard categories first
    # Verify all the columns in ordered_cats are actually in all_cats
    for key in ordered_cats:
        if is_standard_category(key) and key in all_cat_set:
            cat_ord.append(key)
    # Add any new standard cats at the end of the list
    for key in all_cats:
        if key not in cat_ord and is_standard_category(key):
            cat_ord.append(key)
    # Now add the non-standard cats (user cats and search). As these are always
    # hierarchical, only keep the prefix.
    user_cat_prefixes = set()
    for key in all_cats:
        if not is_standard_category(key):
            prefix = key.partition('.')[0]
            if prefix not in user_cat_prefixes:
                cat_ord.append(prefix)
                user_cat_prefixes.add(prefix)
    return cat_ord


numeric_collation = prefs['numeric_collation']


def sort_key_for_popularity(x, hierarchical_categories=None):
    return (-getattr(x, 'count', 0), sort_key(x.sort or x.name))


def sort_key_for_rating(x, hierarchical_categories=None):
    return (-getattr(x, 'avg_rating', 0.0), sort_key(x.sort or x.name))


# When sorting by name, treat the period in hierarchical categories as a tab so
# that the value sorts above similarly-named items. Example: "foo.bar" should
# sort above "foo a.bar". Without this substitution "foo.bar" sorts below "foo
# a.bar" because '.' sorts higher than space.


def sort_key_for_name_and_first_letter(x, hierarchical_categories=()):
    v1 = icu_upper(x.sort or x.name)
    if x.category in hierarchical_categories:
        v1 = v1.replace('.', '\t')
    v2 = v1 or ' '
    # The idea is that '9999999999' is larger than any digit so all digits
    # will sort in front. Non-digits will sort according to their ICU first letter
    c = v2[0]
    return (c if numeric_collation and c.isdigit() else '9999999999', collation_order(v2), sort_key(v1))


def sort_key_for_name(x, hierarchical_categories=()):
    v = x.sort or x.name
    if x.category not in hierarchical_categories:
        return sort_key(v)
    return sort_key(v.replace('.', '\t'))


category_sort_keys = {True: {}, False: {}}
category_sort_keys[True]['popularity'] = category_sort_keys[False]['popularity'] = sort_key_for_popularity
category_sort_keys[True]['rating'] = category_sort_keys[False]['rating'] = sort_key_for_rating
category_sort_keys[True]['name'] = sort_key_for_name_and_first_letter
category_sort_keys[False]['name'] = sort_key_for_name


# Various parts of calibre depend on the order of fields in the returned
# dict being in the default display order: standard fields, custom in alpha order,
# user categories, then saved searches. This works because the backend adds
# custom columns to field metadata in the right order.
def get_categories(dbcache, sort='name', book_ids=None, first_letter_sort=False, uncollapsed_categories=None):
    if sort not in CATEGORY_SORTS:
        raise ValueError('sort ' + sort + ' not a valid value')

    hierarchical_categories = frozenset(dbcache.pref('categories_using_hierarchy', ()))
    fm = dbcache.field_metadata
    book_rating_map = dbcache.fields['rating'].book_value_map
    lang_map = dbcache.fields['languages'].book_value_map

    categories = OrderedDict()
    book_ids = frozenset(book_ids) if book_ids else book_ids
    pm_cache = {}

    def get_metadata(book_id):
        ans = pm_cache.get(book_id)
        if ans is None:
            ans = pm_cache[book_id] = dbcache._get_proxy_metadata(book_id)
        return ans

    bids = None
    uncollapsed_categories = () if uncollapsed_categories is None else uncollapsed_categories

    for category, is_multiple, is_composite in find_categories(fm):
        fl_sort = False if category in uncollapsed_categories else bool(first_letter_sort)
        tag_class = create_tag_class(category, fm)
        sort_on, reverse = sort, False
        if is_composite:
            if bids is None:
                bids = dbcache._all_book_ids() if book_ids is None else book_ids
            cats = dbcache.fields[category].get_composite_categories(tag_class, book_rating_map, bids, is_multiple, get_metadata)
        elif category == 'news':
            cats = dbcache.fields['tags'].get_news_category(tag_class, book_ids)
        else:
            cat = fm[category]
            brm = book_rating_map
            dt = cat['datatype']
            if dt == 'rating':
                if category != 'rating':
                    brm = dbcache.fields[category].book_value_map
                if sort_on == 'name':
                    sort_on, reverse = 'rating', True
            cats = dbcache.fields[category].get_categories(tag_class, brm, lang_map, book_ids)
            if category != 'authors' and dt == 'text' and cat['is_multiple'] and cat['display'].get('is_names', False):
                for item in cats:
                    item.sort = author_to_author_sort(item.sort)
        cats.sort(
            key=partial(category_sort_keys[fl_sort][sort_on], hierarchical_categories=hierarchical_categories),
            reverse=reverse,
        )
        categories[category] = cats

    # Needed for legacy databases that have multiple ratings that
    # map to n stars
    for r in categories['rating']:
        for x in tuple(categories['rating']):
            if r.name == x.name and r.id != x.id:
                r.id_set |= x.id_set
                r.count = len(r.id_set)
                categories['rating'].remove(x)
                break

    # User categories
    user_categories = clean_user_categories(dbcache).copy()

    # First add any grouped search terms to the user categories
    muc = dbcache.pref('grouped_search_make_user_categories', [])
    gst = dbcache.pref('grouped_search_terms', {})
    for c in gst:
        if c not in muc:
            continue
        uc = []
        for sc in gst[c]:
            for t in categories.get(sc, ()):
                uc.append([t.name, sc, 0])
        user_categories[c] = uc

    if user_categories:
        # We want to use same node in the user category as in the source
        # category. To do that, we need to find the original Tag node. There is
        # a time/space tradeoff here. By converting the tags into a map, we can
        # do the verification in the category loop much faster, at the cost of
        # temporarily duplicating the categories lists.
        taglist = {}
        for c, items in categories.items():
            taglist[c] = {icu_lower(t.name): t for t in items}

        # Add the category values to the user categories
        for user_cat in sorted(user_categories, key=sort_key):
            items = []
            names_seen = {}
            user_cat_is_gst = user_cat in gst
            for name, label, ign in user_categories[user_cat]:
                n = icu_lower(name)
                if label in taglist and n in taglist[label]:
                    if user_cat_is_gst:
                        # for gst items, make copy and consolidate the tags by name.
                        if n in names_seen:
                            # We must combine this node into a previous one with
                            # the same name ignoring case. As part of the process,
                            # remember the source categories and correct the
                            # average rating
                            t = names_seen[n]
                            other_tag = taglist[label][n]
                            t.id_set |= other_tag.id_set
                            t.count = len(t.id_set)
                            t.original_categories.add(other_tag.category)

                            total_rating = 0
                            count = 0
                            for id_ in t.id_set:
                                rating = book_rating_map.get(id_, 0)
                                if rating:
                                    total_rating += rating / 2
                                    count += 1
                            if total_rating and count:
                                t.avg_rating = total_rating / count
                        else:
                            # Must deepcopy so we don't share the id_set between nodes
                            t = copy.deepcopy(taglist[label][n])
                            t.original_categories = {t.category}
                            names_seen[n] = t
                            items.append(t)
                    else:
                        items.append(taglist[label][n])
                # else: do nothing, to not include nodes w zero counts
            cat_name = '@' + user_cat  # add the '@' to avoid name collision
            items.sort(key=partial(category_sort_keys[False][sort], hierarchical_categories=hierarchical_categories))
            categories[cat_name] = items

    # ### Finally, the saved searches category ####
    items = []
    queries = dbcache._search_api.saved_searches.queries
    for srch in sorted(queries, key=sort_key):
        items.append(Tag(srch, sort=srch, search_expression=queries[srch], category='search', is_editable=False))
    if items:
        categories['search'] = items

    return categories



# autogenerated by __main__.py do not edit
name_map = {
    'QDBus': 'PyQt6.QtDBus',
    'QDBusAbstractAdaptor': 'PyQt6.QtDBus',
    'QDBusAbstractInterface': 'PyQt6.QtDBus',
    'QDBusArgument': 'PyQt6.QtDBus',
    'QDBusConnection': 'PyQt6.QtDBus',
    'QDBusConnectionInterface': 'PyQt6.QtDBus',
    'QDBusError': 'PyQt6.QtDBus',
    'QDBusInterface': 'PyQt6.QtDBus',
    'QDBusMessage': 'PyQt6.QtDBus',
    'QDBusObjectPath': 'PyQt6.QtDBus',
    'QDBusPendingCall': 'PyQt6.QtDBus',
    'QDBusPendingCallWatcher': 'PyQt6.QtDBus',
    'QDBusPendingReply': 'PyQt6.QtDBus',
    'QDBusReply': 'PyQt6.QtDBus',
    'QDBusServiceWatcher': 'PyQt6.QtDBus',
    'QDBusSignature': 'PyQt6.QtDBus',
    'QDBusUnixFileDescriptor': 'PyQt6.QtDBus',
    'QDBusVariant': 'PyQt6.QtDBus',
    'QtDBus': 'PyQt6.QtDBus',
}
module_names = frozenset(('QtDBus',))



#!/usr/bin/env python
# License: GPLv3 Copyright: 2015, Kovid Goyal <kovid at kovidgoyal.net>

import mimetypes
import os
import re
import subprocess
from collections import defaultdict
from plistlib import loads

from calibre.ptempfile import TemporaryDirectory
from calibre.utils.icu import numeric_sort_key

application_locations = ('/Applications', '~/Applications', '~/Desktop')

# Public UTI MAP {{{


def generate_public_uti_map():
    from urllib.request import urlopen

    from html5_parser import parse
    from lxml import etree

    raw = urlopen(
        'https://developer.apple.com/library/ios/documentation/Miscellaneous/Reference/UTIRef/Articles/System-DeclaredUniformTypeIdentifiers.html'
    ).read()
    root = parse(raw)
    tables = root.xpath('//table')[0::2]
    data = {}
    for table in tables:
        for tr in table.xpath('descendant::tr')[1:]:
            td = tr.xpath('descendant::td')
            identifier = etree.tostring(td[0], method='text', encoding='unicode').strip()
            tags = etree.tostring(td[2], method='text', encoding='unicode').strip()
            identifier = identifier.split()[0].replace('\u200b', '')
            exts = [x.strip()[1:].lower() for x in tags.split(',') if x.strip().startswith('.')]
            for ext in exts:
                data[ext] = identifier
    lines = ['PUBLIC_UTI_MAP = {']
    for ext in sorted(data):
        r = ("'" + ext + "':").ljust(16)
        lines.append((' ' * 4) + r + "'" + data[ext] + "',")
    lines.append('}')
    with open(__file__, 'r+') as f:
        raw = f.read()
        f.seek(0)
        nraw = re.sub(r'^PUBLIC_UTI_MAP = .+?}', '\n'.join(lines), raw, flags=re.MULTILINE | re.DOTALL)
        f.truncate(), f.write(nraw)


# Generated by generate_public_uti_map()
PUBLIC_UTI_MAP = {
    '3g2': 'public.3gpp2',
    '3gp': 'public.3gpp',
    '3gp2': 'public.3gpp2',
    '3gpp': 'public.3gpp',
    'ai': 'com.adobe.illustrator.ai-image',
    'aif': 'public.aiff-audio',
    'aifc': 'public.aifc-audio',
    'aiff': 'public.aiff-audio',
    'app': 'com.apple.application-bundle',
    'applescript': 'com.apple.applescript.text',
    'asf': 'com.microsoft.advanced-systems-format',
    'asx': 'com.microsoft.advanced-stream-redirector',
    'au': 'public.ulaw-audio',
    'avi': 'public.avi',
    'bin': 'com.apple.macbinary-archive',
    'bmp': 'com.microsoft.bmp',
    'bundle': 'com.apple.bundle',
    'c': 'public.c-source',
    'c++': 'public.c-plus-plus-source',
    'caf': 'com.apple.coreaudio-format',
    'cc': 'public.c-plus-plus-source',
    'class': 'com.sun.java-class',
    'command': 'public.shell-script',
    'cp': 'public.c-plus-plus-source',
    'cpio': 'public.cpio-archive',
    'cpp': 'public.c-plus-plus-source',
    'csh': 'public.csh-script',
    'cxx': 'public.c-plus-plus-source',
    'defs': 'public.mig-source',
    'dfont': 'com.apple.truetype-datafork-suitcase-font',
    'dll': 'com.microsoft.windows-dynamic-link-library',
    'doc': 'com.microsoft.word.doc',
    'efx': 'com.js.efx-fax',
    'eps': 'com.adobe.encapsulated-postscript',
    'exe': 'com.microsoft.windows-executable',
    'exp': 'com.apple.symbol-export',
    'exr': 'com.ilm.openexr-image',
    'fpx': 'com.kodak.flashpix.image',
    'framework': 'com.apple.framework',
    'gif': 'com.compuserve.gif',
    'gtar': 'org.gnu.gnu-tar-archive',
    'gz': 'org.gnu.gnu-zip-archive',
    'gzip': 'org.gnu.gnu-zip-archive',
    'h': 'public.c-header',
    'h++': 'public.c-plus-plus-header',
    'hpp': 'public.c-plus-plus-header',
    'hqx': 'com.apple.binhex-archive',
    'htm': 'public.html',
    'html': 'public.html',
    'hxx': 'public.c-plus-plus-header',
    'icc': 'com.apple.colorsync-profile',
    'icm': 'com.apple.colorsync-profile',
    'icns': 'com.apple.icns',
    'ico': 'com.microsoft.ico',
    'jar': 'com.sun.java-archive',
    'jav': 'com.sun.java-source',
    'java': 'com.sun.java-source',
    'javascript': 'com.netscape.javascript-source',
    'jfx': 'com.j2.jfx-fax',
    'jnlp': 'com.sun.java-web-start',
    'jp2': 'public.jpeg-2000',
    'jpeg': 'public.jpeg',
    'jpg': 'public.jpeg',
    'js': 'com.netscape.javascript-source',
    'jscript': 'com.netscape.javascript-source',
    'key': 'com.apple.keynote.key',
    'kth': 'com.apple.keynote.kth',
    'm': 'public.objective-c-source',
    'm15': 'public.mpeg',
    'm4a': 'public.mpeg-4-audio',
    'm4b': 'com.apple.protected-mpeg-4-audio',
    'm4p': 'com.apple.protected-mpeg-4-audio',
    'm75': 'public.mpeg',
    'mdimporter': 'com.apple.metadata-importer',
    'mig': 'public.mig-source',
    'mm': 'public.objective-c-plus-plus-source',
    'mov': 'com.apple.quicktime-movie',
    'mp3': 'public.mp3',
    'mp4': 'public.mpeg-4',
    'mpeg': 'public.mpeg',
    'mpg': 'public.mpeg',
    'o': 'public.object-code',
    'otf': 'public.opentype-font',
    'pct': 'com.apple.pict',
    'pdf': 'com.adobe.pdf',
    'pf': 'com.apple.colorsync-profile',
    'pfa': 'com.adobe.postscript.pfa-font',
    'pfb': 'com.adobe.postscript-pfb-font',
    'ph3': 'public.php-script',
    'ph4': 'public.php-script',
    'php': 'public.php-script',
    'php3': 'public.php-script',
    'php4': 'public.php-script',
    'phtml': 'public.php-script',
    'pic': 'com.apple.pict',
    'pict': 'com.apple.pict',
    'pl': 'public.perl-script',
    'plugin': 'com.apple.plugin',
    'pm': 'public.perl-script',
    'png': 'public.png',
    'pntg': 'com.apple.macpaint-image',
    'ppt': 'com.microsoft.powerpoint.ppt',
    'ps': 'com.adobe.postscript',
    'psd': 'com.adobe.photoshop-image',
    'py': 'public.python-script',
    'qif': 'com.apple.quicktime-image',
    'qt': 'com.apple.quicktime-movie',
    'qtif': 'com.apple.quicktime-image',
    'qtz': 'com.apple.quartz-composer-composition',
    'r': 'com.apple.rez-source',
    'ra': 'com.real.realaudio',
    'ram': 'com.real.realaudio',
    'rb': 'public.ruby-script',
    'rbw': 'public.ruby-script',
    'rm': 'com.real.realmedia',
    'rtf': 'public.rtf',
    'rtfd': 'com.apple.rtfd',
    's': 'public.assembly-source',
    'scpt': 'com.apple.applescript.script',
    'sd2': 'com.digidesign.sd2-audio',
    'sgi': 'com.sgi.sgi-image',
    'sh': 'public.shell-script',
    'sit': 'com.allume.stuffit-archive',
    'sitx': 'com.allume.stuffit-archive',
    'smil': 'com.real.smil',
    'snd': 'public.ulaw-audio',
    'suit': 'com.apple.font-suitcase',
    'tar': 'public.tar-archive',
    'tga': 'com.truevision.tga-image',
    'tgz': 'org.gnu.gnu-zip-tar-archive',
    'tif': 'public.tiff',
    'tiff': 'public.tiff',
    'ttc': 'public.truetype-collection-font',
    'ttf': 'public.truetype-ttf-font',
    'txt': 'public.plain-text',
    'ulw': 'public.ulaw-audio',
    'vcard': 'public.vcard',
    'vcf': 'public.vcard',
    'vfw': 'public.avi',
    'wav': 'com.microsoft.waveform-audio',
    'wave': 'com.microsoft.waveform-audio',
    'wax': 'com.microsoft.windows-media-wax',
    'wdgt': 'com.apple.dashboard-widget',
    'wm': 'com.microsoft.windows-media-wm',
    'wma': 'com.microsoft.windows-media-wma',
    'wmp': 'com.microsoft.windows-media-wmp',
    'wmv': 'com.microsoft.windows-media-wmv',
    'wmx': 'com.microsoft.windows-media-wmx',
    'wvx': 'com.microsoft.windows-media-wvx',
    'xbm': 'public.xbitmap-image',
    'xls': 'com.microsoft.excel.xls',
    'xml': 'public.xml',
    'zip': 'com.pkware.zip-archive',
}
PUBLIC_UTI_RMAP = defaultdict(set)
for ext, uti in PUBLIC_UTI_MAP.items():
    PUBLIC_UTI_RMAP[uti].add(ext)
PUBLIC_UTI_RMAP = dict(PUBLIC_UTI_RMAP)

# }}}


def find_applications_in(base):
    try:
        entries = os.listdir(base)
    except OSError:
        return
    for name in entries:
        path = os.path.join(base, name)
        if os.path.isdir(path):
            if name.lower().endswith('.app'):
                yield path
            else:
                yield from find_applications_in(path)


def find_applications():
    for base in application_locations:
        base = os.path.expanduser(base)
        yield from find_applications_in(base)


def get_extensions_from_utis(utis, plist):
    declared_utis = defaultdict(set)
    for key in ('UTExportedTypeDeclarations', 'UTImportedTypeDeclarations'):
        for decl in plist.get(key, ()):
            if isinstance(decl, dict):
                uti = decl.get('UTTypeIdentifier')
                if isinstance(uti, (str, bytes)):
                    spec = decl.get('UTTypeTagSpecification')
                    if isinstance(spec, dict):
                        ext = spec.get('public.filename-extension')
                        if ext:
                            declared_utis[uti] |= set(ext)
                        types = spec.get('public.mime-type')
                        if types:
                            for mt in types:
                                for ext in mimetypes.guess_all_extensions(mt, strict=False):
                                    declared_utis[uti].add(ext.lower()[1:])
    ans = set()
    for uti in utis:
        ans |= declared_utis[uti]
        ans |= PUBLIC_UTI_RMAP.get(uti, set())
    return ans


def get_bundle_data(path):
    path = os.path.abspath(path)
    info = os.path.join(path, 'Contents', 'Info.plist')
    ans = {
        'name': os.path.splitext(os.path.basename(path))[0],
        'path': path,
    }
    try:
        with open(info, 'rb') as f:
            plist = loads(f.read())
    except Exception:
        import traceback

        traceback.print_exc()
        return None
    ans['name'] = plist.get('CFBundleDisplayName') or plist.get('CFBundleName') or ans['name']
    icfile = plist.get('CFBundleIconFile')
    if icfile:
        icfile = os.path.join(path, 'Contents', 'Resources', icfile)
        if not os.path.exists(icfile):
            icfile += '.icns'
        if os.path.exists(icfile):
            ans['icon_file'] = icfile
    bid = plist.get('CFBundleIdentifier')
    if bid:
        ans['identifier'] = bid
    ans['extensions'] = extensions = set()
    for dtype in plist.get('CFBundleDocumentTypes', ()):
        utis = frozenset(dtype.get('LSItemContentTypes', ()))
        if utis:
            extensions |= get_extensions_from_utis(utis, plist)
        else:
            for ext in dtype.get('CFBundleTypeExtensions', ()):
                if isinstance(ext, (str, bytes)):
                    extensions.add(ext.lower())
            for mt in dtype.get('CFBundleTypeMIMETypes', ()):
                if isinstance(mt, (str, bytes)):
                    for ext in mimetypes.guess_all_extensions(mt, strict=False):
                        extensions.add(ext.lower())
    return ans


def find_programs(extensions):
    extensions = frozenset(extensions)
    ans = []
    for app in find_applications():
        try:
            app = get_bundle_data(app)
        except Exception:
            import traceback

            traceback.print_exc()
            continue
        if app and app['extensions'].intersection(extensions):
            ans.append(app)
    return ans


def get_icon(path, pixmap_to_data=None, as_data=False, size=64):
    if not path:
        return
    with TemporaryDirectory() as tdir:
        iconset = os.path.join(tdir, 'output.iconset')
        try:
            subprocess.check_call(['iconutil', '-c', 'iconset', '-o', 'output.iconset', path], cwd=tdir)
        except subprocess.CalledProcessError:
            return
        try:
            names = os.listdir(iconset)
        except OSError:
            return
        if not names:
            return
        from qt.core import QImage, Qt

        names.sort(key=numeric_sort_key)
        for name in names:
            m = re.search(r'(\d+)x\d+', name)
            if m is not None and int(m.group(1)) >= size:
                ans = QImage(os.path.join(iconset, name))
                if not ans.isNull():
                    break
        else:
            return
    ans = ans.scaled(size, size, transformMode=Qt.TransformationMode.SmoothTransformation)
    if as_data:
        assert pixmap_to_data is not None
        ans = pixmap_to_data(ans)
    return ans


def entry_to_cmdline(entry, path):
    app = entry['path']
    if os.path.isdir(app):
        return ['open', '-a', app, path]
    if 'identifier' in entry:
        return ['open', '-b', entry['identifier'], path]
    return [app, path]



# License: GPLv3 Copyright: 2011, John Schember <john@nachtimwald.com>

import re
from functools import partial
from random import shuffle

from qt.core import QCheckBox, QDialog, QDialogButtonBox, QGridLayout, QIcon, QLabel, QMenu, QSize, QStyle, Qt, QTabWidget, QTimer, QVBoxLayout, QWidget

from calibre.gui2 import JSONConfig, error_dialog, info_dialog
from calibre.gui2.dialogs.choose_format import ChooseFormatDialog
from calibre.gui2.ebook_download import show_download_info
from calibre.gui2.progress_indicator import ProgressIndicator
from calibre.gui2.store.config.chooser.chooser_widget import StoreChooserWidget
from calibre.gui2.store.config.search.search_widget import StoreConfigWidget
from calibre.gui2.store.search.adv_search_builder import AdvSearchBuilderDialog
from calibre.gui2.store.search.download_thread import CacheUpdateThreadPool, SearchThreadPool
from calibre.gui2.store.search.models import Matches
from calibre.gui2.store.search.search_ui import Ui_Dialog
from calibre.gui2.widgets2 import HistoryLineEdit2
from calibre.utils.filenames import ascii_filename
from calibre.utils.localization import _


def add_items_to_context_menu(widget: HistoryLineEdit2, menu: QMenu) -> None:
    menu.addSeparator()
    ac = menu.addAction(_('Clear search &history'))
    assert ac is not None
    ac.triggered.connect(widget.clear_history)


class SearchDialog(QDialog, Ui_Dialog):
    SEARCH_TEXT = _('&Search')
    STOP_TEXT = _('&Stop')

    def __init__(self, gui, parent=None, query=''):
        QDialog.__init__(self, parent)
        self.setupUi(self)
        s = self.style()
        assert s is not None
        self.close_button.setIcon(s.standardIcon(QStyle.StandardPixmap.SP_DialogCloseButton))

        self.config = JSONConfig('store/search')
        self.search_title.initialize('store_search_search_title')
        self.search_author.initialize('store_search_search_author')
        self.search_edit.initialize('store_search_search')
        self.search_title.add_items_to_context_menu_callback = partial(add_items_to_context_menu, self.search_title)
        self.search_author.add_items_to_context_menu_callback = partial(add_items_to_context_menu, self.search_author)
        self.search_edit.add_items_to_context_menu_callback = partial(add_items_to_context_menu, self.search_edit)

        # Loads variables that store various settings.
        # This needs to be called soon in __init__ because
        # the variables it sets up are used later.
        self.load_settings()

        self.gui = gui

        # Setup our worker threads.
        self.search_pool = SearchThreadPool(self.search_thread_count)
        self.cache_pool = CacheUpdateThreadPool(self.cache_thread_count)
        _rv_model = self.results_view.model()
        assert _rv_model is not None
        assert isinstance(_rv_model, Matches)
        _rv_model.cover_pool.set_thread_count(self.cover_thread_count)
        _rv_model.details_pool.set_thread_count(self.details_thread_count)
        self.results_view.setCursor(Qt.CursorShape.PointingHandCursor)
        # needed for live updates of amazon_live.py
        from calibre.live import start_worker

        start_worker()

        # Check for results and hung threads.
        self.checker = QTimer()
        self.progress_checker = QTimer()
        self.hang_check = 0

        # Update store caches silently.
        for p in self.gui.istores.values():
            self.cache_pool.add_task(p, self.timeout)

        self.store_checks = {}
        self.setup_store_checks()

        # Set the search query
        if isinstance(query, (bytes, str)):
            self.search_edit.setText(query)
        elif isinstance(query, dict):
            if 'author' in query:
                self.search_author.setText(query['author'])
            if 'title' in query:
                self.search_title.setText(query['title'])

        # Create and add the progress indicator
        self.pi = ProgressIndicator(self, 24)
        self.button_layout.takeAt(0)
        self.button_layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.button_layout.insertWidget(0, self.pi, 0, Qt.AlignmentFlag.AlignCenter)

        self.adv_search_button.setIcon(QIcon.ic('gear.png'))
        self.adv_search_button.setToolTip(_('Advanced search'))
        self.configure.setIcon(QIcon.ic('config.png'))

        self.adv_search_button.clicked.connect(self.build_adv_search)
        self.search.clicked.connect(self.toggle_search)
        self.search_title.returnPressed.connect(self.toggle_search)
        self.search_author.returnPressed.connect(self.toggle_search)
        self.search_edit.returnPressed.connect(self.toggle_search)
        self.checker.timeout.connect(self.get_results)
        self.progress_checker.timeout.connect(self.check_progress)
        self.results_view.activated.connect(self.result_item_activated)
        self.results_view.download_requested.connect(self.download_book)
        self.results_view.open_requested.connect(self.open_store)
        _rv_model.total_changed.connect(self.update_book_total)
        self.select_all_stores.clicked.connect(self.stores_select_all)
        self.select_invert_stores.clicked.connect(self.stores_select_invert)
        self.select_none_stores.clicked.connect(self.stores_select_none)
        self.configure.clicked.connect(self.do_config)
        self.finished.connect(self.dialog_closed)
        self.searching = False

        self.progress_checker.start(100)

        self.restore_state()

    def setup_store_checks(self):
        first_run = self.config.get('first_run', True)

        # Add check boxes for each store so the user
        # can disable searching specific stores on a
        # per search basis.
        existing = {}
        for n in self.store_checks:
            existing[n] = self.store_checks[n].isChecked()

        self.store_checks = {}

        stores_check_widget = QWidget()
        store_list_layout = QGridLayout()
        stores_check_widget.setLayout(store_list_layout)

        icon = QIcon.ic('donate.png')
        for i, x in enumerate(sorted(self.gui.istores.keys(), key=lambda x: x.lower())):
            cbox = QCheckBox(x)
            cbox.setChecked(existing.get(x, first_run))
            store_list_layout.addWidget(cbox, i, 0, 1, 1)
            if self.gui.istores[x].base_plugin.affiliate:
                iw = QLabel(self)
                iw.setToolTip('<p>' + _('Buying from this store supports the calibre developer: %s</p>') % self.gui.istores[x].base_plugin.author + '</p>')
                iw.setPixmap(icon.pixmap(16, 16))
                store_list_layout.addWidget(iw, i, 1, 1, 1)
            self.store_checks[x] = cbox
        store_list_layout.setRowStretch(store_list_layout.rowCount(), 10)
        self.store_list.setWidget(stores_check_widget)

        self.config['first_run'] = False

    def build_adv_search(self):
        adv = AdvSearchBuilderDialog(self)
        if adv.exec() == QDialog.DialogCode.Accepted:
            self.search_edit.setText(adv.search_string())

    def resize_columns(self):
        total = 600
        # Cover
        self.results_view.setColumnWidth(0, 85)
        total = total - 85
        # Title / Author
        self.results_view.setColumnWidth(1, int(total * 0.40))
        # Price
        self.results_view.setColumnWidth(2, int(total * 0.12))
        # DRM
        self.results_view.setColumnWidth(3, int(total * 0.15))
        # Store / Formats
        self.results_view.setColumnWidth(4, int(total * 0.25))
        # Download
        self.results_view.setColumnWidth(5, 20)
        # Affiliate
        self.results_view.setColumnWidth(6, 20)

    def toggle_search(self):
        if self.searching:
            self.search_pool.abort()
            m = self.results_view.model()
            assert m is not None
            assert isinstance(m, Matches)
            m.details_pool.abort()
            m.cover_pool.abort()
            self.search.setText(self.SEARCH_TEXT)
            self.checker.stop()
            self.searching = False
        else:
            self.do_search()
        # Prevent hitting the enter key twice in quick succession causing
        # the search to start and stop
        self.search.setEnabled(False)
        QTimer.singleShot(1000, lambda: self.search.setEnabled(True))

    def do_search(self):
        # Stop all running threads.
        self.checker.stop()
        self.search_pool.abort()
        # Clear the visible results.
        _m = self.results_view.model()
        assert _m is not None
        assert isinstance(_m, Matches)
        _m.clear_results()

        # Don't start a search if there is nothing to search for.
        query = []
        if self.search_title.text():
            query.append('title2:"~{}"'.format(str(self.search_title.text()).replace('"', ' ')))
        if self.search_author.text():
            query.append('author2:"{}"'.format(str(self.search_author.text()).replace('"', ' ')))
        if self.search_edit.text():
            query.append(str(self.search_edit.text()))
        query = ' '.join(query)
        if not query.strip():
            error_dialog(self, _('No query'), _('You must enter a title, author or keyword to search for.'), show=True)
            return
        self.searching = True
        self.search.setText(self.STOP_TEXT)
        # Give the query to the results model so it can do
        # further filtering.
        _m.set_query(query)

        # Plugins are in random order that does not change.
        # Randomize the order of the plugin names every time
        # there is a search. This way plugins closer
        # to a don't have an unfair advantage over
        # plugins further from a.
        store_names = list(self.store_checks)
        if not store_names:
            return
        # Remove all of our internal filtering logic from the query.
        query = self.clean_query(query)
        shuffle(store_names)
        # Add plugins that the user has checked to the search pool's work queue.
        self.gui.istores.join(4.0)  # Wait for updated plugins to load
        for n in store_names:
            if self.store_checks[n].isChecked():
                self.search_pool.add_task(query, n, self.gui.istores[n], self.max_results, self.timeout)
        self.hang_check = 0
        self.checker.start(100)
        self.pi.startAnimation()

    def clean_query(self, query):
        query = query.lower()
        # Remove control modifiers.
        query = query.replace('\\', '')
        query = query.replace('!', '')
        query = query.replace('=', '')
        query = query.replace('~', '')
        query = query.replace('>', '')
        query = query.replace('<', '')
        # Remove the prefix.
        for loc in ('all', 'author', 'author2', 'authors', 'title', 'title2'):
            query = re.sub(rf'{loc}:"(?P<a>[^\s"]+)"', r'\g<a>', query)
            query = query.replace(f'{loc}:', '')
        # Remove the prefix and search text.
        for loc in ('cover', 'download', 'downloads', 'drm', 'format', 'formats', 'price', 'store'):
            query = re.sub(rf'{loc}:"[^"]"', '', query)
            query = re.sub(rf'{loc}:[^\s]*', '', query)
        # Remove logic.
        query = re.sub(r'(^|\s|")(and|not|or|a|the|is|of)(\s|$|")', ' ', query)
        # Remove "
        query = query.replace('"', '')
        # Remove excess whitespace.
        query = re.sub(r'\s+', ' ', query)
        query = query.strip()
        return query.encode('utf-8')

    def save_state(self):
        self.save_geometry(self.config, 'geometry')
        self.config['store_splitter_state'] = bytearray(self.store_splitter.saveState())
        rv_model = self.results_view.model()
        assert rv_model is not None
        assert isinstance(rv_model, Matches)
        self.config['results_view_column_width'] = [self.results_view.columnWidth(i) for i in range(rv_model.columnCount())]
        self.config['sort_col'] = rv_model.sort_col
        self.config['sort_order'] = rv_model.sort_order.value
        self.config['open_external'] = self.open_external.isChecked()

        store_check = {}
        for k, v in self.store_checks.items():
            store_check[k] = v.isChecked()
        self.config['store_checked'] = store_check

    def restore_state(self):
        self.restore_geometry(self.config, 'geometry')
        splitter_state = self.config.get('store_splitter_state', None)
        if splitter_state:
            self.store_splitter.restoreState(splitter_state)

        results_cwidth = self.config.get('results_view_column_width', None)
        if results_cwidth:
            for i, x in enumerate(results_cwidth):
                rv_model2 = self.results_view.model()
                assert rv_model2 is not None
                if i >= rv_model2.columnCount():
                    break
                self.results_view.setColumnWidth(i, x)
        else:
            self.resize_columns()

        self.open_external.setChecked(self.should_open_external)

        store_check = self.config.get('store_checked', None)
        if store_check:
            for n in store_check:
                if n in self.store_checks:
                    self.store_checks[n].setChecked(store_check[n])

        _rv_m = self.results_view.model()
        assert _rv_m is not None
        assert isinstance(_rv_m, Matches)
        _rv_m.sort_col = self.config.get('sort_col', 2)
        so = self.config.get('sort_order', Qt.SortOrder.AscendingOrder)
        if isinstance(so, int):
            so = Qt.SortOrder(so)
        _rv_m.sort_order = so
        rv_header = self.results_view.header()
        assert rv_header is not None
        rv_header.setSortIndicator(_rv_m.sort_col, so)

    def load_settings(self):
        # Seconds
        self.timeout = self.config.get('timeout', 75)
        # Milliseconds
        self.hang_time = self.config.get('hang_time', 75) * 1000

        self.max_results = self.config.get('max_results', 15)
        self.should_open_external = self.config.get('open_external', True)

        # Number of threads to run for each type of operation
        self.search_thread_count = self.config.get('search_thread_count', 4)
        self.cache_thread_count = self.config.get('cache_thread_count', 2)
        self.cover_thread_count = self.config.get('cover_thread_count', 2)
        self.details_thread_count = self.config.get('details_thread_count', 4)

    def do_config(self):
        # Save values that need to be synced between the dialog and the
        # search widget.
        self.config['open_external'] = self.open_external.isChecked()

        # Create the config dialog. It's going to put two config widgets
        # into a QTabWidget for displaying all of the settings.
        d = QDialog(self)
        button_box = QDialogButtonBox(QDialogButtonBox.StandardButton.Close)
        v = QVBoxLayout(d)
        button_box.accepted.connect(d.accept)
        button_box.rejected.connect(d.reject)
        d.setWindowTitle(_('Customize Get books search'))

        tab_widget = QTabWidget(d)
        v.addWidget(tab_widget)
        v.addWidget(button_box)

        chooser_config_widget = StoreChooserWidget()
        search_config_widget = StoreConfigWidget(self.config)

        tab_widget.addTab(chooser_config_widget, _('Choose s&tores'))
        tab_widget.addTab(search_config_widget, _('Configure s&earch'))

        # Restore dialog state.
        d.restore_geometry(self.config, 'config_dialog_geometry')
        tab_index = self.config.get('config_dialog_tab_index', 0)
        tab_index = min(tab_index, tab_widget.count() - 1)
        tab_widget.setCurrentIndex(tab_index)

        d.exec()

        # Save dialog state.
        d.save_geometry(self.config, 'config_dialog_geometry')
        self.config['config_dialog_tab_index'] = tab_widget.currentIndex()

        search_config_widget.save_settings()
        self.config_changed()
        self.gui.load_store_plugins()
        self.setup_store_checks()

    def sizeHint(self):
        return QSize(800, 600)

    def config_changed(self):
        self.load_settings()

        self.open_external.setChecked(self.should_open_external)
        self.search_pool.set_thread_count(self.search_thread_count)
        self.cache_pool.set_thread_count(self.cache_thread_count)
        _cc_m = self.results_view.model()
        assert _cc_m is not None
        assert isinstance(_cc_m, Matches)
        _cc_m.cover_pool.set_thread_count(self.cover_thread_count)
        _cc_m.details_pool.set_thread_count(self.details_thread_count)

    def get_results(self):
        # We only want the search plugins to run
        # a maximum set amount of time before giving up.
        self.hang_check += 1
        if self.hang_check >= self.hang_time:
            self.search_pool.abort()
            self.checker.stop()
        # Stop the checker if not threads are running.
        elif not self.search_pool.threads_running() and not self.search_pool.has_tasks():
            self.checker.stop()

        _gr_m = self.results_view.model()
        assert _gr_m is not None
        assert isinstance(_gr_m, Matches)
        while self.search_pool.has_results():
            res, store_plugin = self.search_pool.get_result()
            if res:
                _gr_m.add_result(res, store_plugin)

        if not self.search_pool.threads_running() and not _gr_m.has_results():
            info_dialog(
                self,
                _('No matches'),
                _("Couldn't find any books matching your query."),
                show=True,
                show_copy_button=False,
            )

    def update_book_total(self, total):
        self.total.setText(f'{total}')

    def result_item_activated(self, index):
        _ria_m = self.results_view.model()
        assert _ria_m is not None
        assert isinstance(_ria_m, Matches)
        result = _ria_m.get_result(index)

        if result.downloads:
            self.download_book(result)
        else:
            self.open_store(result)

    def download_book(self, result):
        d = ChooseFormatDialog(self, _('Choose format to download to your library.'), list(result.downloads.keys()))
        if d.exec() == QDialog.DialogCode.Accepted:
            ext = d.format()
            fname = result.title[:60] + '.' + ext.lower()
            fname = ascii_filename(fname)
            show_download_info(result.title, parent=self)
            self.gui.download_ebook(result.downloads[ext], filename=fname, create_browser=result.create_browser)

    def open_store(self, result):
        self.gui.istores[result.store_name].open(self, result.detail_item, self.open_external.isChecked())

    def check_progress(self):
        m = self.results_view.model()
        assert m is not None
        assert isinstance(m, Matches)
        if not self.search_pool.threads_running() and not m.cover_pool.threads_running() and not m.details_pool.threads_running():
            self.pi.stopAnimation()
            self.search.setText(self.SEARCH_TEXT)
            self.searching = False
        else:
            self.searching = True
            if str(self.search.text()) != self.STOP_TEXT:
                self.search.setText(self.STOP_TEXT)
            if not self.pi.isAnimated():
                self.pi.startAnimation()

    def stores_select_all(self):
        for check in self.store_checks.values():
            check.setChecked(True)

    def stores_select_invert(self):
        for check in self.store_checks.values():
            check.setChecked(not check.isChecked())

    def stores_select_none(self):
        for check in self.store_checks.values():
            check.setChecked(False)

    def dialog_closed(self, result):
        _dc_m = self.results_view.model()
        assert _dc_m is not None
        assert isinstance(_dc_m, Matches)
        _dc_m.closing()
        self.search_pool.abort()
        self.cache_pool.abort()
        self.save_state()

    def exec(self):
        if str(self.search_edit.text()).strip() or str(self.search_title.text()).strip() or str(self.search_author.text()).strip():
            self.do_search()
        return QDialog.exec(self)

    exec_ = exec


if __name__ == '__main__':
    import sys

    from calibre.gui2 import Application
    from calibre.gui2.preferences.main import init_gui

    app = Application([])
    app
    gui = init_gui()

    s = SearchDialog(gui, query=' '.join(sys.argv[1:]))
    s.exec()



# License: GPLv3 Copyright: 2022, Vaso Peras-Likodric <vaso at vipl.in.rs>

import re
from typing import ClassVar

from calibre.devices.kindle.apnx_page_generator.generators.fast_page_generator import FastPageGenerator
from calibre.devices.kindle.apnx_page_generator.i_page_generator import IPageGenerator, mobi_html
from calibre.devices.kindle.apnx_page_generator.pages import Pages


class PagebreakPageGenerator(IPageGenerator):
    instance: ClassVar[PagebreakPageGenerator]

    def name(self) -> str:
        return 'pagebreak'

    def _generate_fallback(self, mobi_file_path: str, real_count: int | None) -> Pages:
        return FastPageGenerator.instance.generate(mobi_file_path, real_count)

    def _generate(self, mobi_file_path: str, real_count: int | None) -> Pages:
        """Determine pages based on the presence of <*pagebreak*/>."""
        html = mobi_html(mobi_file_path)
        pages = []
        for m in re.finditer(rb'<[^>]*pagebreak[^>]*>', html):
            pages.append(m.end())

        return Pages(pages)


PagebreakPageGenerator.instance = PagebreakPageGenerator()



# License: GPLv3 Copyright: 2011, John Schember <john@nachtimwald.com>

from qt.core import QDialog, QDialogButtonBox, QVBoxLayout

from calibre.gui2.store.config.chooser.chooser_widget import StoreChooserWidget
from calibre.utils.localization import _


class StoreChooserDialog(QDialog):
    def __init__(self, parent):
        QDialog.__init__(self, parent)

        self.setWindowTitle(_('Choose stores'))

        button_box = QDialogButtonBox(QDialogButtonBox.StandardButton.Close)
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)
        v = QVBoxLayout(self)
        self.config_widget = StoreChooserWidget()
        v.addWidget(self.config_widget)
        v.addWidget(button_box)

        self.resize(800, 600)



#!/usr/bin/env python
# License: GPLv3 Copyright: 2008, Kovid Goyal kovid@kovidgoyal.net

"""
Read metadata from LRX files
"""

import struct
from zlib import decompress

from calibre.ebooks.metadata import MetaInformation, string_to_authors
from calibre.utils.xml_parse import safe_xml_fromstring


def _read(f, at, amount):
    f.seek(at)
    return f.read(amount)


def word_be(buf):
    return struct.unpack('>L', buf)[0]


def word_le(buf):
    return struct.unpack('<L', buf)[0]


def short_le(buf):
    return struct.unpack('<H', buf)[0]


def short_be(buf):
    return struct.unpack('>H', buf)[0]


def get_metadata(f):
    def read(at, amount):
        return _read(f, at, amount)

    f.seek(0)
    buf = f.read(12)
    if buf[4:] == b'ftypLRX2':
        offset = 0
        while True:
            offset += word_be(buf[:4])
            try:
                buf = read(offset, 8)
            except Exception:
                raise ValueError('Not a valid LRX file')
            if buf[4:] == b'bbeb':
                break
        offset += 8
        buf = read(offset, 16)
        if buf[:8].decode('utf-16-le') != 'LRF\x00':
            raise ValueError('Not a valid LRX file')
        lrf_version = word_le(buf[8:12])
        offset += 0x4C
        compressed_size = short_le(read(offset, 2))
        offset += 2
        if lrf_version >= 800:
            offset += 6
        compressed_size -= 4
        uncompressed_size = word_le(read(offset, 4))
        info = decompress(f.read(compressed_size))
        if len(info) != uncompressed_size:
            raise ValueError('LRX file has malformed metadata section')
        root = safe_xml_fromstring(info)
        bi = root.find('BookInfo')
        assert bi is not None
        title = bi.find('Title')
        assert title is not None
        title_sort = title.get('reading', None)
        title = title.text
        author = bi.find('Author')
        assert author is not None
        author_sort = author.get('reading', None)
        mi = MetaInformation(title, string_to_authors(author.text))
        mi.title_sort, mi.author_sort = title_sort, author_sort
        author = author.text
        publisher = bi.find('Publisher')
        mi.publisher = getattr(publisher, 'text', None)
        mi.tags = [x.text for x in bi.findall('Category')]
        docinfo = root.find('DocInfo')
        assert docinfo is not None
        lang_elem = docinfo.find('Language')
        assert lang_elem is not None
        mi.language = lang_elem.text
        return mi

    elif buf[4:8] == b'LRX':
        raise ValueError('Librie LRX format not supported')
    else:
        raise ValueError('Not a LRX file')



# License: GPLv3 Copyright: 2008, Kovid Goyal <kovid at kovidgoyal.net>

from PIL import ImageFont

'''
Default fonts used in the PRS500
'''

LIBERATION_FONT_MAP = {
    'Swis721 BT Roman': 'LiberationSans-Regular',
    'Dutch801 Rm BT Roman': 'LiberationSerif-Regular',
    'Courier10 BT Roman': 'LiberationMono-Regular',
}

FONT_FILE_MAP = {}


def get_font(name, size, encoding='unic'):
    """
    Get an ImageFont object by name.
    @param size: Font height in pixels. To convert from pts:
                 sz in pixels = (dpi/72) * size in pts
    @param encoding: Font encoding to use. E.g. 'unic', 'symbol', 'ADOB', 'ADBE', 'aprm'
    @param manager: A dict that will store the PersistentTemporary
    """
    from calibre.utils.resources import get_path as P

    if name in LIBERATION_FONT_MAP:
        return ImageFont.truetype(P(f'fonts/liberation/{LIBERATION_FONT_MAP[name]}.ttf'), size, encoding=encoding)
    elif name in FONT_FILE_MAP:
        return ImageFont.truetype(FONT_FILE_MAP[name], size, encoding=encoding)



#!/usr/bin/env python
# License: GPLv3 Copyright: 2013, Kovid Goyal <kovid at kovidgoyal.net>


class Settings:
    def __init__(self, namespace):
        self.default_tab_stop = 720 / 20
        self.namespace = namespace

    def __call__(self, root):
        for dts in self.namespace.XPath('//w:defaultTabStop[@w:val]')(root):
            try:
                self.default_tab_stop = int(self.namespace.get(dts, 'w:val')) / 20
            except ValueError, TypeError, AttributeError:
                pass
